#
# SciNet's DAT112, Neural Network Programming.
# Lecture 8, 19 May 2022.
# Erik Spence.
#
# This file, MNIST_gan.py, contains code used for lecture 8.  It is a
# script which trains a generative adversarial network on the MNIST
# dataset.
#

#######################################################################


"""
MNIST_gan.py contains a script which trains a generative
adversarial network on the MNIST dataset.  This code was inspired by
many online examples, but the main one was
https://www.tensorflow.org/tutorials/generative/dcgan

"""

#######################################################################


import tensorflow as tf
import tensorflow.keras.models as km
import tensorflow.keras.layers as kl
import tensorflow.keras.utils as ku

import matplotlib.pyplot as plt
import numpy as np
import os

import numpy.random as npr


#######################################################################


## This code is needed to run on Mist, the SciNet GPU cluster.  In
## most situations it can probably be removed.
gpus = tf.config.list_physical_devices('GPU')
for gpu in gpus:
    tf.config.experimental.set_memory_growth(gpu, True)


######################################################################


## Size of the latent space.
noise_dim = 100

## Number of epochs to train.
EPOCHS = 500

## The batch size to train.
BATCH_SIZE = 256

## The number of data to use for training.
BUFFER_SIZE = 60000

    
######################################################################


print('Loading data')

mnist_path = os.getenv('HOME') + '/.keras/datasets/mnist.npz'
(train_images, train_labels), _ = tf.keras.datasets.mnist.load_data(mnist_path)
train_images = train_images.reshape(train_images.shape[0], 28,
                                    28, 1).astype('float32')

## Normalize the images to (-1, 1)
train_images = (train_images - 127.5) / 127.5 

## Create a Tensorflow Dataset.
train_dataset = tf.data.Dataset.from_tensor_slices(train_images).shuffle(BUFFER_SIZE).batch(BATCH_SIZE)


######################################################################


def make_generator_model():

    model = km.Sequential(name = 'G')

    model.add(kl.Dense(7 * 7 * 256, input_shape=(100,)))
    model.add(kl.BatchNormalization())
    model.add(kl.LeakyReLU())

    model.add(kl.Reshape((7, 7, 256)))

    model.add(kl.Conv2DTranspose(256, (5, 5), strides=(1, 1),
                                 padding='same'))
    model.add(kl.BatchNormalization())
    model.add(kl.LeakyReLU())

    model.add(kl.Conv2DTranspose(128, (5, 5), strides=(1, 1),
                                 padding='same'))
    model.add(kl.BatchNormalization())
    model.add(kl.LeakyReLU())

    model.add(kl.Conv2DTranspose(64, (5, 5), strides=(2, 2),
                                 padding='same'))
    model.add(kl.BatchNormalization())
    model.add(kl.LeakyReLU())

    model.add(kl.Conv2DTranspose(32, (5, 5), strides=(2, 2),
                                 padding='same'))
    model.add(kl.BatchNormalization())
    model.add(kl.LeakyReLU())

    model.add(kl.Conv2DTranspose(1, (5, 5), strides=(1, 1),
                                 padding='same',
                                 activation='tanh'))

    return model


######################################################################


def make_discriminator_model():

    model = km.Sequential(name = 'D')

    model.add(kl.Conv2D(32, (5, 5), strides=(1, 1), padding='same',
                        input_shape=[28, 28, 1]))
    model.add(kl.LeakyReLU())
    model.add(kl.Dropout(0.3))

    model.add(kl.Conv2D(64, (5, 5), strides=(2, 2), padding='same'))
    model.add(kl.LeakyReLU())
    model.add(kl.Dropout(0.3))

    model.add(kl.Conv2D(128, (5, 5), strides=(2, 2), padding='same'))
    model.add(kl.LeakyReLU())
    model.add(kl.Dropout(0.3))

    model.add(kl.Conv2D(256, (5, 5), strides=(1, 1), padding='same'))
    model.add(kl.LeakyReLU())
    model.add(kl.Dropout(0.3))

    model.add(kl.Flatten())
    model.add(kl.Dense(2, activation = 'softmax'))

    model.compile(optimizer = tf.keras.optimizers.Adam(1e-4),
                  loss = 'categorical_crossentropy',
                  metrics = ['accuracy'])
    
    return model


####################################################################

                    
def plot_images(generator, noise, step = 0):

    """
    This function creates a 4 x 4 grid of image plots, based on the
    input noise and the desired clases.
    
    Inputs: 

    - generator: NN generator model, used to generate the images.
   
    - noise: a (16, 100) matrix of input noise.
    
    - classes: a 16-element vector if input digits, 0-9.

    - step: the step index, added to the output file name.

    Outputs: None.

    """

    if (not os.path.isdir('data')): os.makedirs('data')

    # The filename into which the image is saved.
    filename = "data/mnist_%05d.png" % step

    # Create the images.
    images = generator.predict(noise)

    # Create an empty figure.
    plt.figure(figsize = (10, 10))

    # For each image...
    for i in range(images.shape[0]):

        # Change to the next subplot.
        plt.subplot(4, 4, i + 1)

        # Rescale the image back to its 0-255 form.
        image = images[i, :, :, :] * 127.5 + 127.5

        # Reshape to remove the extra dimensions.
        image = np.reshape(image, [28, 28])

        # Plot the image.
        plt.imshow(image, cmap = 'gray_r')

        # Plot the digit that that particular image should be.
        # plt.text(23, 26, str(classes[i][0]), fontsize = 30, color = 'r')

        # Turn off the axis.
        plt.axis('off')

    # Tighten up the boundaries, and save.
    plt.tight_layout()
    plt.savefig(filename)
    plt.close('all')
    

######################################################################


if __name__ == "__main__":

    print('building generator')
    G = make_generator_model()

    print('building discriminator')
    D = make_discriminator_model()

    ## Create the input layer for the Generator.
    input_z = kl.Input(shape = (noise_dim,))

    ## Create the full model, to train the Generator, and compile the
    ## model.
    output_class = D(G(inputs = input_z))
    DG = km.Model(inputs = input_z, outputs = output_class)

    DG.get_layer('D').trainable = False

    DG.compile(optimizer = tf.keras.optimizers.Adam(1e-4),
               loss = 'categorical_crossentropy')


    ## These data are used for the images we will save to make the
    ## animation.
    noise_input = npr.normal(0., 1.0, size = [16, noise_dim])
    
    D_losses = []
    DG_losses = []


    ## Let us iterate!
    for it in range(EPOCHS):

        for image_batch in train_dataset:             

            # unfreeze D
            D.trainable = True
            for layer in D.layers: layer.trainable = True

            ## Now create the fake images, based on the current state of
            ## the generator.
            zz = tf.random.normal([BATCH_SIZE, noise_dim]) 
            generated_images = G.predict(zz)

            ## Bundle all the images together.
            all_images = np.concatenate([generated_images,
                                         image_batch])

            ## Bundle all the categories (real/fake) together.
            all_cats = np.concatenate([np.zeros(BATCH_SIZE),
                                       np.ones(image_batch.shape[0])])
            all_cats = ku.to_categorical(all_cats, 2)

            ## Train the Discriminator.
            D_loss = D.train_on_batch(all_images, all_cats)

            # Save the loss.
            D_losses.append(D_loss)

            ## We're done training the discriminator, for this
            ## iteration.  Now train the generator.

            ## Freeze D.
            D.trainable = False
            for layer in D.layers: layer.trainable = False

            ## Create some noise, and fake classes.
            zz = tf.random.normal([BATCH_SIZE, noise_dim]) 

            ## Now train the combined DG on the generated data, but with
            ## only training G.  Note that we are setting the category to
            ## be 'real', so that the generator will create images that
            ## the discriminator thinks are real.
            DG_loss = DG.train_on_batch(zz,
                                        ku.to_categorical(np.ones(BATCH_SIZE),2))

            ## Again, save the losses.
            DG_losses.append(DG_loss)

        ## Every epoch print out the losses.
        log_mesg = "%d: [D loss: %f]" % (it, D_losses[-1][0])
        log_mesg = "%s  [DG loss: %f]" % (log_mesg, DG_losses[-1])
        print(log_mesg)

        plot_images(G, noise_input, step = it)


######################################################################
